function DuoToneM
%
% LIGO LSC end-to-end timing diagnostic module
%
% Operation is based on dual-tone sinusoids at frequencies around 1kHz and
% exactly 1Hz diference in frequency. The lower frequency line should be
% sitting on a 60Hz harmonic due to LSC Pentek cross talk issues.
%
% The script has no command line input/output. All communication happens
% through files:
%
% Input: 
%     ./DuoToneTmpData.dat -    int16 big endian binary file 3 * L points
%           + This data file must contain two byte integers (int16)
%           + There must be data from three channels strictly in the
%           following order: 
%               & Dual tone sinusoid from the LSC crate
%               & Dual tone sinusoid from the PEM crate
%               & Ramp (1PPS) sinal  from the PEM crate
%           + Each channel must contain exactly L seconds of data
%           + This file is read from the local directory (./)
%
% Output: 
%     ./Results.txt - plane text file containing the double precision results 
%           + Time of common zero phase for LSC sinusoids ('zSlsc') (line 1) 
%           + Time of zero crossing phase for PEM ramp    ('zRpem') (line 2) 
%           + Time of common zero phase for PEM sinusoids ('zSpem') (line 3)
%           + Time delay of the LSC timestamp to 1PPS     ('zTot') (line 4)
%
% This code should be compiled using the:
%       >> mcc -R -nojvm -R -nojit -mv DuoToneM
% command on fortress.ligo-wa.caltech.edu.
% 
% The compiled code will be DuoToneM and DuoToneM.ctf, which must be
% transferred to the same directory as the DuoTone executable.
%
% The envitonment executing this compiled matlab code must be set right:
%
% setenv LD_LIBRARY_PATH ${LD_LIBRARY_PATH}:/export/home/smarka/MCR/v72/runtime/sol2:
%                                           /export/home/smarka/MCR/v72/sys/os/sol2:
%                                           /export/home/smarka/MCR/v72/sys/java/jre/sol2/jre1.5.0/lib/sparc/native_threads:
%                                           /export/home/smarka/MCR/v72/sys/java/jre/sol2/jre1.5.0/lib/sparc/client:
%                                           /export/home/smarka/MCR/v72/sys/java/jre/sol2/jre1.5.0/lib/sparc:
%                                           /opt/dmt/new/lib:./:/export/home/smarka/DuoTone/ 
%
% setenv XAPPLRESDIR /export/home/smarka/MCR/X11/app-defaults
%
% Szabi Marka, 2005
% v0.1 11/06/2005


% tic
%load('DuoTonePars.mat'); % in case a faster convergence is needed ... load the previous results.

format long g;
warning off all;

L = 40; % Each data segment is 40 sec long...
Fs=16384; % Sampling rate of channels
T = ( ( 1:1:Fs ) - 1 ) / Fs; % Time axis
T = T(:)';

% Load input data
fid = fopen('DuoToneTmpData.dat', 'r'); 
dSlsc = fread(fid, 16384*L, 'int16', 'b');
% figure(111);
% plot(dSlsc);
[ mSlsc, sSlsc, mi ] = Squeeze(dSlsc, Fs, L, 0);
dSpem = fread(fid, 16384*L, 'int16', 'b');
% figure(112);
% plot(dSpem);
[ mSpem, sSpem, mi ] = Squeeze(dSpem, Fs, L, 0);
dRpem = fread(fid, 16384*L, 'int16', 'b');
% figure(113);
% plot(dRpem);
[ mRpem, sRpem, mi ] = Squeeze(dRpem, Fs, L, 1);
fclose(fid);

% ---------

% Seed parameters for the fit
% 1 - Normalized amplitude ( larger sinusoid)
% 2 - Time delay of common zero phase within the second
betaLSC = [ 0.57 0 ];
betaPEM = [ 0.57 0 ];

% Do the simultaneous constrained fits for the PEM and LSC sinusoids
betaPEM = nlinfit(T,mSpem,@SinSinFit, betaPEM(:), statset('TolFun',1e-5,'TolX', 1e-5, 'MaxIter', 50));
zSpem = 1e6 * betaPEM(2);

betaLSC = nlinfit(T,mSlsc,@SinSinFit, betaLSC(:), statset('TolFun',1e-5,'TolX', 1e-5, 'MaxIter', 50));
zSlsc = 1e6 * betaLSC(2);

% save('DuoTonePars.mat', 'betaLSC', 'betaPEM');

% figure(3);
% plot(mSpem,'b');
% hold on;
% plot(100000*sSpem,'y.');
% hold off;
% ---------

% Extract the zero crossing of the PEM 1PPS ramp
T = ((1:1:3*Fs) - (Fs + 1)) / Fs;
b = round(mi + Fs/2 - 15);
e = b + Fs -1;
mR = [ mRpem mRpem mRpem ];
sR = [ sRpem sRpem sRpem ];

mR = mR(b:e);
sR = sR(b:e);
T  = T(b:e);

[m, mi] = max(mR);
mi = min(mi);

i = mi - 31 + find(abs(mR(mi-30:mi) - 0.45) < 0.41);
i = i(min(2,length(i)):min(10,length(i)));

p = polyfit(T(i), mR(i), 1);
L = polyval(p,T(i));
RR = L./mR(i);
RR = RR / prod(RR);
mR(i) = RR(:)' .* mR(i);

p = polyfit(mR(i), T(i), 1);
zRpem = polyval(p,0) * 1e6; % microseconds

% figure(1);
% L = polyval(p,[-0.1:0.001:1.1]);
% plot(T(8180:8220),mR(8180:8220),'b');
% hold on;
% plot(T(8180:8220),1625*sR(8180:8220),'c');
% plot(T(i),mR(i),'rp');
% plot(L,[-0.1:0.001:1.1],'m');
% hold off;

zTot = zSlsc - ( zSpem - zRpem );

% Seve the result
save('Results.txt', 'zSlsc', 'zRpem', 'zSpem', 'zTot', '-ascii', '-double');
%toc

return;

% -------------------------

% Dual tone sinusoids to fit with
function Y = SinSinFit(beta,T)

Y1 = (        beta(1) * sin(2 * pi * 960 * ( T - beta(2) ) ) );
Y2 = ( 0.75 * beta(1) * sin(2 * pi * 961 * ( T - beta(2) ) ) );

Y = Y1+Y2;

return;

% -------------------------

% Take the 40 1 second long segments and pile them on top of each other
% Compute the statistics
function [ m, s, mi ] = Squeeze(d, Fs, L, f)

if ( f == 0 )
    d = d(:)' - median(d);
end
d = reshape(d, Fs, L)';
m = mean(d);
if ( f == 1 )
   [ma, mi] = max(m);
   so = sort(m);
   m = m - mean(so(512:Fs-512)); 
end
[ma, mi] = max(m);
m = m / ma;
s = std(d) / ma;

return;

