/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef KLEINEWELLEN_HH
#define KLEINEWELLEN_HH

#include "MultiStream.hh"
#include "FilterDesign.hh"
#include "Pipe.hh"
#include "TSeries.hh"
#include "autopipe.hh"
#include <fstream>
#include <string>
#include <iosfwd>
#include <cstdio>
#include <vector>
#include <list>

class LPEFilter;

struct kwVeto {
    kwVeto(const std::string& cf) : conffile(cf), nfactor(0) {}
    std::string conffile;
    int         nfactor;
};

struct Element {
    Element(double energy, double normEnergy, int index, int scale,
        double startTime, double endTime, double dt, int clusterID)
    : energy(energy), normEnergy(normEnergy), index(index), scale(scale),
      startTime(startTime), endTime(endTime), dt(dt), clusterID(clusterID)
    {}
    int dist(const Element& b) const;
    bool operator<(const Element& b) const {return startTime < b.startTime;}
    double energy;
    double normEnergy;
    int index;
    int scale;
    double startTime;
    double endTime;
    double dt;
    int clusterID;
};

inline int
Element::dist(const Element& b) const {
    int dscale = scale - b.scale;
    if (dscale <= 0) return -dscale + abs((index   >> -dscale) - b.index);
    else             return  dscale + abs((b.index >>  dscale) -   index);
}

typedef std::vector<Element>   element_vect;
typedef element_vect::iterator element_iter;

struct Event {
    Event(const std::string& name);
    bool operator<(const Event& b) const {return startTime < b.startTime;}
    Event& operator += (const Element& el);
    void average(const Time& toff, double fs);
    double startTime;
    double endTime;
    double centralTime;
    double frequency;
    double energy;
    double normEnergy;
    int npts;
    double significance;
    std::string channelName;
};

typedef std::vector<Event>   event_vect;
typedef event_vect::iterator event_iter;

// negative natural logarithm of significance
double
logSignificance(double normalizedEnergy, double timeFrequencyArea,
        double precision);

class kwChan {
public:
    /**  Channel processor constructor.
      */
    kwChan(const std::string& name);

    /**  Copy constructor
     */
    kwChan(const kwChan& x) {*this = x;}

    /**  Channel processor destructor.
      */
    ~kwChan(void);

    /**  Assignment operator
     */
    kwChan& operator=(const kwChan& x);

    /**  Print out the channel parameters.
     */
    std::ostream& dump_params(std::ostream& out) const;

    /**  Process a data stride.
     */
    event_vect process(MultiDacc& dacc);

    /**  Set up the sample rate dependent filtering.
      */
    void setup(const TSeries* ts);

public:                        // public options that need to be set from outside
    std::string channel;       // channel name of data (noise)
    float    lowFrequency;     // low frequency boundary
    float    highFrequency;    // high frequency boundary
    double   significance;     // significance threshold for reporting clusters
    double   threshold;        // gaussian threshold for black pixels
    int      maxDist;          // max distance between two black pixels
    int      decimateFactor;   // number of decimations to do
    Interval transientDuration; // amount of data to skip at boundaries

private:
    int      fs;               // sample rate (after any decimations)
    int      lowScale;         // frequency bounds converted into wavelet scale
    int      highScale;
    int      size;             // size of the wavelet decomposition array
    int      lpefLength;
    Wavelet*     wavePtr;      // pointer to wavelet functions library
    auto_pipe    prefilter;    // decimator and highpass prefilter
    LPEFilter*   lpefilter;    // linear predictive whitening filter
    TSeries      tsLpef;       // linear predictive filtered timeseries

};

class kleineWelle : public MultiStream {
public:

    kleineWelle(int argc, const char *argv[]);
    ~kleineWelle(void);

    void ProcessData(void);

private:
    Interval stride;     // number of seconds to analyze at a time

    std::ifstream optionsFile;
    typedef std::list<kwChan>      channel_list;
    typedef channel_list::iterator channel_iter;
    typedef std::list<kwVeto>      vetoseg_list;
    typedef vetoseg_list::iterator vetoseg_iter;
    channel_list mChanList;
    vetoseg_list    mVetoList;
    std::string basename;
    Time lasttime;       // start time of previous anlaysis segment
};

std::ostream&
operator<<(std::ostream& output, const Event& a);

#endif // KLEINEWELLE_HH 
