/* -*-  mode: c++; c-basic-offset: 3; -*- */
#ifndef SEGLIST_HH
#define SEGLIST_HH

#include "LockSegment.hh"
#include <stdexcept>
#include <vector>

//======================================  Segment list class
/**  LockSegList is a list of LockSegment objects ordered by Start Time.
  *
  *  @memo Segment list class.
  *  @author J.Zweizig
  *  @version 1.2; last modified 11/03/2006 
  */
class LockSegList {
public:
   typedef unsigned long size_type;

public:
   /** Segment list constructor.
     */
   explicit LockSegList(const char* lid, const char* file=0);

   /** Segment list destructor.
     */
   virtual ~LockSegList(void) {}


   /** Segment default constructor.
     */
   LockSegList(void) {}

   /**  Check that segments are inserted in order... print errors to cerr.
    */
   bool check(void) const;

   /**  Erase all the segments in the list.
    */
   void clear(void);

   /**  Coalesce overlapping segments.
     */
   void coalesce(void);

   /**  Look for the segment containing the specified time.
     */
   size_type find(const Time& t) const throw(std::domain_error);

   /**  Find the first segment that ends after the specified time. The index
     *  of the found segment is returned. If there are no segments after the 
     *  specified time, an index equal to the size of segment list is returned.
     */ 
   size_type findafter(const Time& t) const;

   /** Return the list title.
     */
   const char* getListID(void) const;

   /**  Test whether the specified time is within one of the segments in the
     *  list.
     */
   bool inSegment(const Time& t) const;

   /**  Test whether any time between the specified start and end time is
     *  within one or more of the list segments.
     */
   bool inSegment(const Time& t, const Time& t2) const;

   /**  Insert a copy of the specified segment into the list. The segment
     *  is inserted to maintain an increasing start-time order.
     */
   void insert(const LockSegment& l);

   /**  Invert the segment list. The gaps between successive segments are 
     *  converted to segments and the original segments are removed.
     */
   void invert(void);

   /**  @memo Calculate the "live time" of the segments list.
     *  Return the sum of all segment durations between the start and the 
     *  end time.
     */
   Interval live(Time start=Time(0), Time end=Time(SEGMENT_TINF)) const;

   /**  Insert a segment into the list. If the segment overlaps existing
     *  one or more segments, the segments are combined.
     *  @merge a segment into the list.
     *  @param seg Segent to be merged intot the list.
     */
   void merge(const LockSegment& seg);

   /**  Pad the front and back of each segment in the list by pFront and 
     *  pBack. The pFront and pBack intervals are defined such that a positive 
     *  interval increases the segment duration and a negative interval 
     *  decreases the resulting segment duration. i.e. the resulting segment 
     *  is #[start - pStart, end + pBack]#. Segments are coalesced after the 
     *  the padding is added.
     *  @param pFront Length of padding to add to the start of the segment.
     *  @param pBack  Length of padding to add to the back of the segment.
     */
   void pad(Interval pFront=0, Interval pBack=0);

   /**  Fill segment list from the contents of the specified file.
     */
   void read(const std::string& file);

   /**  Set the debug printout level.
     */
   void setDebug(int lvl);

   /**  Return the number of segments in the list.
     */
   size_type size(void) const;

   /**  Sort the list.
     */
   void sort(void);

   /**  Add an segment to the end of the list (unsorted insertion).
     */
   void stuff(const LockSegment& seg);

   /**  Return a constant reference to the specified segment.
     */
   const LockSegment& operator[](size_type seg) const;

   /**  Return a constant reference to the segment containing the specified
     *  time.
     */
   const LockSegment& operator[](const Time& t) const throw(std::domain_error);

   /**  Replace the current list with the union of the list and the argument.
     */
   LockSegList& operator|=(const LockSegList& t);

   /**  Replace the current list with the intersection of the list and the 
     *  argument.
     */
   LockSegList& operator&=(const LockSegList& t);

   /**  Replace the current list with the differences between the current 
     *  list and the argument list argument.
     */
   LockSegList& operator^=(const LockSegList& t);

   /**  Test whether any segments in the list overlap the specified time span.
    */
   bool overlap(const Time& t, const Time& t2) const;

   /**  Write a segment specifier to the list. The list specifier consists 
     *	of the List ID followed by a hyphen and the segment identifier.
     */
   std::ostream& putID(std::ostream& out, size_type inx) const;

private:
   LockSegment& ref(size_type seg);
   const LockSegment& ref(size_type seg) const;
   void checkhdr(void) const;
   void stretch(void);

public:
   typedef std::vector<LockSegment> SegList_type;
   typedef SegList_type::const_iterator seg_iter;

private:
   std::string  mListID;
   SegList_type mList;
   int          mDebug;
};

//======================================  Segment list inlined methods
inline const char*
LockSegList::getListID(void) const {
   return mListID.c_str();
}

inline const LockSegment&
LockSegList::operator[](size_type seg) const {
   return mList[seg];
}

inline const LockSegment&
LockSegList::operator[](const Time& t) const throw(std::domain_error) {
   return mList[find(t)];
}

inline LockSegment&
LockSegList::ref(size_type seg) {
   return mList[seg];
}

inline const LockSegment&
LockSegList::ref(size_type seg) const {
   return mList[seg];
}

inline LockSegList::size_type
LockSegList::size(void) const {
   return mList.size();
}

#endif  // !defined(SEGLIST_HH)
