/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "sockutil.h"
#include "lsmp_con.hh"
#include "lsmp_prod.hh"
#include "SigFlag.hh"
#include <string>
#include <stdexcept>
#include <inttypes.h>
#include <iostream>
#include <cstdlib>
#include <errno.h>

/**  The FrameLink program connects two shared memory partitions together 
  *  with a tcp link. The program may run in one of four modes:
  *  \begin{itemize}
  *  \item push: establish a connection to a receive process and send 
  *              it frames
  */
class FrameLink {
public:
    enum pgm_mode {
	kUnknown,
	kPush,
	kPull,
	kReceive,
	kSend
    };
public:
    FrameLink(int argc, const char* argv[]);
    ~FrameLink(void);
    bool client(void) const;
    void process(void);

    /**  Read the specified number of bytes into the buffer pointed to by
      *  buf. Returns number of bytes read, 0 on eof, or -1.
      */
    int read(void* buf, int len) const;
    bool receiver(void) const;
    bool sender(void) const;
    bool server(void) const;
    void stats(void);
private:
    //----------------------------------  COmmand line parameters
    pgm_mode    mPgmMode;
    std::string mPeerName;
    std::string mPartition;
    int         mDebug;

    //----------------------------------  Status info.
    SigFlag     mTerm;
    int         mServer;
    int         mSocket;
    LSMP_CON*   mInPart;
    LSMP_PROD*  mOutPart;
    bool        mConnected;
    long        mBlocks;
};

inline bool
FrameLink::client(void) const {
    return (mPgmMode == kPush || mPgmMode == kPull);
}

inline bool
FrameLink::receiver(void) const {
    return (mPgmMode == kReceive || mPgmMode == kPull);
}

inline bool
FrameLink::sender(void) const {
    return (mPgmMode == kSend || mPgmMode == kPush);
}

inline bool
FrameLink::server(void) const {
    return (mPgmMode == kSend || mPgmMode == kReceive);
}

struct link_header {
    uint32_t block_id;
    uint32_t length;
    uint32_t gps_secs;
    uint32_t data_time;
};

using namespace std;
static const double Timeout = 60;

//======================================  Main program
int
main(int argc, const char* argv[]) {
    try {
	FrameLink pgm(argc, argv);
	pgm.process();
    } catch (exception& e) {
	cerr << "framelink: caught exception: " << e.what() << endl;
    }
}

//======================================  Initialization
FrameLink::FrameLink(int argc, const char* argv[]) 
  : mPgmMode(kUnknown), mDebug(0), mServer(-1),  mSocket(-1), mInPart(0), 
    mOutPart(0), mConnected(false), mBlocks(0)
{
    int iarg    = 1;
    bool syntax = false;
    for (iarg=1; iarg < argc; iarg++) {
	if (*argv[iarg] != '-') break;
	string argi = argv[iarg];
	if (argi == "--debug") {
	    mDebug = strtol(argv[++iarg], 0, 0);
	} else {
	    syntax = true;
	    break;
	}
    }
    if (argc-iarg < 3 || syntax) {
	cerr << "framelink command syntax: " << endl;
	cerr << "  framelink <mode> <from> <to>" << endl;
	cerr << "where: " << endl;
	cerr << "<mode> is the program mode: push, pull, send or receive" 
	     << endl;
	cerr << "<from> is the data source partition or socket" << endl;
	cerr << "<to>   is the data destination partition or socket" << endl;
	throw runtime_error("Command syntax error");
    }

    //----------------------------------  Get the program mode.
    string mode = argv[iarg];
    if (mode == "push") {
	mPgmMode   = kPush;
	mPartition = argv[iarg+1];
	mPeerName  = argv[iarg+2];
    } else if (mode == "pull") {
	mPgmMode   = kPull;
	mPeerName  = argv[iarg+1];
	mPartition = argv[iarg+2];
    } else if (mode == "send") {
	mPgmMode   = kSend;
	mPartition = argv[iarg+1];
	mPeerName  = argv[iarg+2];
    } else if (mode == "receive") {
	mPgmMode   = kReceive;
	mPeerName  = argv[iarg+1];
	mPartition = argv[iarg+2];
    } else {
	throw runtime_error(string("Invalid program mode: ") + mode);
    }

    //----------------------------------  Open the partition
    switch (mPgmMode) {
    case kPush:
    case kSend:
	mInPart  = new LSMP_CON(mPartition.c_str());
	break;
    case kPull:
    case kReceive:
	mOutPart = new LSMP_PROD(mPartition.c_str());
	break;
    default:
	break;
    }

    //----------------------------------  Get the socket address
    sockaddr_in sockpeer;
    sockpeer.sin_family      = AF_INET;
    sockpeer.sin_port        = 23232;
    string::size_type i = mPeerName.find(":");
    if (i != string::npos) {
	sockpeer.sin_port = strtol(mPeerName.c_str()+i+1, 0, 0);
	mPeerName.erase(i);
    }
    if (mDebug) {
	cout << "Look up peer: " << mPeerName << ":" << sockpeer.sin_port 
	     << endl;
    }
    int rc = nslookup(mPeerName.c_str(), &sockpeer.sin_addr);
    if (rc < 0) {
	perror("framelink: nslookup error");
	throw runtime_error(string("Can't find address for: ") + mPeerName);
    }

    //----------------------------------  Open a socket
    mSocket = socket(PF_INET, SOCK_STREAM, 0);
    if (mSocket < 0) {
	perror("framelink: Error getting socket");
	throw runtime_error("Socket request failed");
    }

    //----------------------------------  Bind/connect the socket

    if (client()) {
	//------------------------------  Client sockets need connection.
	sockaddr_in socknam;
	socknam.sin_family      = AF_INET;
	socknam.sin_port        = 0;
	socknam.sin_addr.s_addr = INADDR_ANY;
	int len                 = sizeof(socknam);
	if (bind(mSocket, (struct sockaddr *)&socknam, len) < 0) {
	    perror("Failed to bin to client port");
	    throw runtime_error("client port bind failed");
	}

	rc = connectWithTimeout(mSocket, (sockaddr*)&sockpeer, 
				    sizeof(sockpeer), Timeout);
	if (rc) {
	    throw runtime_error(string("Connection failed to: ") + mPeerName);
	}
	mConnected = true;

    } else if (server()) {
	//------------------------------  bind server sockets to specified port
	mServer = mSocket;
	mSocket = -1;
	if (bind(mServer, (struct sockaddr *)&sockpeer, sizeof(sockpeer)) < 0) {
	    perror("Failed to bin to server port");
	    throw runtime_error("server port bind failed");
	}
	if (listen(mServer, 1)) {
	    perror("framelink: Server listen error");
	    throw runtime_error("Listen error");
	}
    }

    //----------------------------------  Catch termintaion signals
    mTerm.add(SIGTERM);
    mTerm.add(SIGINT);
}

//======================================  Destructor
FrameLink::~FrameLink(void) {
    if (mSocket >= 0) close(mSocket);
    if (mServer >= 0) close(mServer);
    if (mInPart)  delete mInPart;
    if (mOutPart) delete mOutPart;
}

//======================================  Copy frames
void
FrameLink::process(void) {
    while (!mTerm) {

	//------------------------------  Wait for a connection if necessary
	while (!mConnected && !mTerm) {
	    if (client()) throw runtime_error("Yikes!");
	    mSocket = accept(mServer, 0, 0);
	    if (mSocket < 0) {
		perror("framelink: Accept error");
	    } else {
		mConnected = true;
	    }
	}
	if (mTerm) continue;

	//------------------------------  Receiver processing...
	if (receiver()) {
	    //--------------------------  Wait for a packet
	    int rc = socketWait(mSocket, Timeout, wm_read);
	    if (rc <= 0) throw runtime_error("Socket error");

	    //--------------------------  Read in the header
	    link_header h;
	    rc = read(&h, sizeof(h));
	    if (rc < 0) {
		throw runtime_error("Error reading header");
	    } else if (!rc) {
		if (server()) {
		    mConnected = false;
		    continue;
		} else {
		    throw runtime_error("end-of-file on read");
		}
	    }

	    if (mDebug) {
		cout << "received header block, len, gps, time = {" 
		     << h.block_id << ", " << h.length << ", " 
		     << h.gps_secs << ", "<< h.data_time << "}" << endl;
	    }

	    //--------------------------  Get a buffer from the output partition
	    uint32_t bufl = mOutPart->getBufferLength();
	    if (h.length > bufl) {
		throw runtime_error("frame length > partition buffer length");
	    }
	    char* buf = mOutPart->get_buffer();
	    if (!buf) {
		throw runtime_error("No buffer available");
	    }

	    //--------------------------  Read the data and release the buffer
	    if (read(buf, h.length) <= 0) {
		mOutPart->return_buffer();
		throw runtime_error("Error receiving frame");
	    }
	    mBlocks++;
	    mOutPart->SetID(h.gps_secs);
	    mOutPart->release(h.length);

	//------------------------------  Sender processing...
	} else {
	    //--------------------------  Get a full frame buffer.
	    link_header h;
	    const char* buf = mInPart->get_buffer();
	    if (!buf) {
		throw runtime_error("Unable to get input buffer");
	    }

	    //--------------------------  Build a header, send it to the peer
	    h.block_id  = mBlocks++;
	    h.length    = mInPart->getLength();
	    h.gps_secs  = mInPart->getEvtID();
	    h.data_time = 1;
	    int rc = send(mSocket, &h, sizeof(h), 0);
	    if (rc != int(sizeof(h))) {
		if (rc < 0) perror("framelink: Error sending header");
		mInPart->free_buffer();
		throw runtime_error("Error sending header");
	    }

	    if (mDebug) {
		cout << "sent header block, len, gps, time = {" << h.block_id
		     << ", " << h.length << ", " << h.gps_secs << ", "
		     << h.data_time << "}" << endl;
	    }

	    //--------------------------  Send the frame data in one swell foop
	    rc = send(mSocket, buf, h.length, 0);
	    if (rc != int(h.length)) {
		if (rc < 0) perror("framelink: Error sending frame");
		mInPart->free_buffer();
		throw runtime_error("Error sending frame");
	    }
	    mInPart->free_buffer();
	}
    }
}

//=======================================  Read the specified length data
int
FrameLink::read(void* buf, int len) const {
    int lread = 0;
    char* bptr = reinterpret_cast<char*>(buf);
    while (lread < len) {
	int rc = recv(mSocket, bptr, len-lread, 0);
	if (rc < 0) {
	    if (errno == EAGAIN) continue;
	    perror("framelink: read error");
	    return -1;
	} if (rc == 0) {
	    break; 
	}
	lread += rc;
	bptr  += rc;
    }
    return lread;
}
